{-# LANGUAGE OverloadedStrings #-}

-- Copyright 2022-2024 Vergara Technologies LLC
-- 
-- This file is part of Zcash-Haskell.
--
-- |
-- Module      :  ZcashHaskell.Orchard
-- Copyright   :  2022-2024 Vergara Technologies
-- License     :  MIT
--
-- Maintainer  :  rene@vergara.network
-- Stability   :  experimental
-- Portability :  unknown
--
-- Functions to interact with the Orchard shielded pool of the Zcash blockchain.
--
module ZcashHaskell.Orchard where

import C.Zcash
  ( rustWrapperGenOrchardReceiver
  , rustWrapperGenOrchardSpendKey
  , rustWrapperOrchardCheck
  , rustWrapperOrchardNoteDecode
  , rustWrapperOrchardNoteDecodeSK
  , rustWrapperReadOrchardCommitmentTree
  , rustWrapperReadOrchardPosition
  , rustWrapperReadOrchardWitness
  , rustWrapperUADecode
  , rustWrapperUfvkDecode
  , rustWrapperUpdateOrchardWitness
  )
import qualified Data.ByteString as BS
import qualified Data.ByteString.Char8 as C
import Data.HexString (HexString(..), fromRawBytes, toBytes)
import qualified Data.Text as T
import qualified Data.Text.Encoding as E
import Data.Word
import Foreign.Rust.Marshall.Variable
import ZcashHaskell.Types
import ZcashHaskell.Utils (encodeBech32, encodeBech32m, f4Jumble)

-- | Derives an Orchard spending key for the given seed and account ID
genOrchardSpendingKey ::
     Seed -- ^ The cryptographic seed for the wallet
  -> CoinType -- ^ The coin type constant
  -> AccountId -- ^ The index of the account to be used
  -> Maybe OrchardSpendingKey
genOrchardSpendingKey s coinType accountId =
  if BS.length k /= 32
    then Nothing
    else Just $ OrchardSpendingKey k
  where
    k =
      withPureBorshVarBuffer $
      rustWrapperGenOrchardSpendKey
        (getBytes s)
        (getValue coinType)
        (fromIntegral accountId)

-- | Derives an Orchard receiver for the given spending key and index
genOrchardReceiver ::
     Int -- ^ The index of the address to be created
  -> Scope -- ^ `External` for wallet addresses, `Internal` for change addresses
  -> OrchardSpendingKey -- ^ The spending key
  -> Maybe OrchardReceiver
genOrchardReceiver i scope osk =
  if BS.length k /= 43
    then Nothing
    else Just $ OrchardReceiver k
  where
    k =
      withPureBorshVarBuffer $
      rustWrapperGenOrchardReceiver
        (getBytes osk)
        (fromIntegral i)
        (scope == External)

-- | Checks if given bytestring is a valid encoded unified address
isValidUnifiedAddress :: BS.ByteString -> Maybe UnifiedAddress
isValidUnifiedAddress str =
  case raw_net decodedAddress of
    0 -> Nothing
    _ -> Just $ makeUA decodedAddress
  where
    decodedAddress = (withPureBorshVarBuffer . rustWrapperUADecode) str
    whichNet =
      case raw_net decodedAddress of
        1 -> MainNet
        2 -> TestNet
        3 -> RegTestNet
    makeUA x =
      UnifiedAddress
        whichNet
        (if BS.length (raw_o x) == 43
           then Just $ OrchardReceiver (raw_o x)
           else Nothing)
        (if BS.length (raw_s x) == 43
           then Just $ SaplingReceiver (raw_s x)
           else Nothing)
        (if not (BS.null (raw_t x))
           then Just $ TransparentReceiver P2PKH (fromRawBytes $ raw_t x)
           else if not (BS.null (raw_to x))
                  then Just $ TransparentReceiver P2SH (fromRawBytes $ raw_to x)
                  else Nothing)

-- | Encode a 'UnifiedAddress' per [ZIP-316](https://zips.z.cash/zip-0316)
encodeUnifiedAddress :: UnifiedAddress -> T.Text
encodeUnifiedAddress ua = encodeBech32m (E.encodeUtf8 hr) b
  where
    hr =
      case ua_net ua of
        MainNet -> uniPaymentAddressHrp
        TestNet -> uniTestPaymentAddressHrp
    b = f4Jumble $ tReceiver <> sReceiver <> oReceiver <> padding
    tReceiver =
      case t_rec ua of
        Nothing -> BS.empty
        Just t ->
          case tr_type t of
            P2SH -> packReceiver 0x01 $ Just $ toBytes $ tr_bytes t
            P2PKH -> packReceiver 0x00 $ Just $ toBytes $ tr_bytes t
    sReceiver = packReceiver 0x02 $ getBytes <$> s_rec ua
    oReceiver = packReceiver 0x03 $ getBytes <$> o_rec ua
    padding = E.encodeUtf8 $ T.justifyLeft 16 '\NUL' hr
    packReceiver :: Word8 -> Maybe BS.ByteString -> BS.ByteString
    packReceiver typeCode receiver' =
      case receiver' of
        Just receiver ->
          if BS.length receiver > 1
            then BS.singleton typeCode `BS.append`
                 (BS.singleton . toEnum . BS.length) receiver `BS.append`
                 receiver
            else BS.empty
        Nothing -> BS.empty

-- | Attempts to decode the given bytestring into a Unified Full Viewing Key
decodeUfvk :: BS.ByteString -> Maybe UnifiedFullViewingKey
decodeUfvk str =
  case net decodedKey of
    0 -> Nothing
    _ -> Just decodedKey
  where
    decodedKey = (withPureBorshVarBuffer . rustWrapperUfvkDecode) str

-- | Check if the given UVK matches the UA given
matchOrchardAddress :: BS.ByteString -> BS.ByteString -> Bool
matchOrchardAddress = rustWrapperOrchardCheck

-- | Attempts to decode the given @OrchardAction@ using the given @UnifiedFullViewingKey@.
decryptOrchardAction ::
     UnifiedFullViewingKey -> OrchardAction -> Maybe DecodedNote
decryptOrchardAction key encAction =
  case a_value decodedAction of
    0 -> Nothing
    _ -> Just decodedAction
  where
    decodedAction =
      withPureBorshVarBuffer $
      rustWrapperOrchardNoteDecode (o_key key) encAction

getSaplingFromUA :: BS.ByteString -> Maybe T.Text
getSaplingFromUA uadd = do
  let a = isValidUnifiedAddress uadd
  case a of
    Nothing -> Nothing
    Just a -> do
      let sraw = s_rec a
      case sraw of
        Nothing -> Nothing
        Just sraw -> do
          let net = ua_net a
          case net of
            MainNet ->
              Just $ encodeBech32 (C.pack sapPaymentAddressHrp) (getBytes sraw)
            TestNet ->
              Just $
              encodeBech32 (C.pack sapTestPaymentAddressHrp) (getBytes sraw)

-- | Attemtps to decode the given @OrchardAction@ using the given @OrchardSpendingKey@
decryptOrchardActionSK ::
     OrchardSpendingKey -> Scope -> OrchardAction -> Maybe DecodedNote
decryptOrchardActionSK sk scope oa =
  case a_value decodedAction of
    0 -> Nothing
    _ -> Just decodedAction
  where
    decodedAction =
      withPureBorshVarBuffer $
      rustWrapperOrchardNoteDecodeSK (getBytes sk) oa (scope == External)

-- | Update a Orchard commitment tree
updateOrchardCommitmentTree ::
     OrchardCommitmentTree -- ^ the base tree
  -> HexString -- ^ the new note commitment
  -> Maybe OrchardCommitmentTree
updateOrchardCommitmentTree tree cmx =
  if BS.length (hexBytes updatedTree) > 1
    then Just $ OrchardCommitmentTree updatedTree
    else Nothing
  where
    updatedTree =
      withPureBorshVarBuffer $
      rustWrapperReadOrchardCommitmentTree
        (hexBytes $ orchTree tree)
        (hexBytes cmx)

-- | Get the Orchard incremental witness from a commitment tree
getOrchardWitness :: OrchardCommitmentTree -> Maybe OrchardWitness
getOrchardWitness tree =
  if BS.length (hexBytes wit) > 1
    then Just $ OrchardWitness wit
    else Nothing
  where
    wit =
      withPureBorshVarBuffer $
      rustWrapperReadOrchardWitness (hexBytes $ orchTree tree)

-- | Get the Sapling note position from a witness
getOrchardNotePosition :: OrchardWitness -> Integer
getOrchardNotePosition =
  fromIntegral . rustWrapperReadOrchardPosition . hexBytes . orchWit

updateOrchardWitness :: OrchardWitness -> [HexString] -> OrchardWitness
updateOrchardWitness wit cmus =
  OrchardWitness $
  withPureBorshVarBuffer $
  rustWrapperUpdateOrchardWitness (toBytes $ orchWit wit) (map toBytes cmus)
