-- Copyright 2022-2024 Vergara Technologies LLC
-- 
-- This file is part of Zcash-Haskell.
--
-- |
-- Module      :  ZcashHaskell.Transparent
-- Copyright   :  2022-2024 Vergara Technologies
-- License     :  MIT
--
-- Maintainer  :  pitmutt@vergara.tech
-- Stability   :  experimental
-- Portability :  unknown
--
-- Functions to interact with the transparent addresses in the Zcash blockchain
--
module ZcashHaskell.Transparent where

import Control.Exception (throwIO)
import Crypto.Hash
import Crypto.Secp256k1
import qualified Data.ByteArray as BA
import qualified Data.ByteString as BS
import Data.ByteString.Base58 (bitcoinAlphabet, decodeBase58, encodeBase58)
import qualified Data.ByteString.Char8 as BC
import Data.Char (chr)
import Data.HexString
import qualified Data.Text as T
import qualified Data.Text.Encoding as E
import Data.Word
import Haskoin.Address (Address(..))
import qualified Haskoin.Crypto.Hash as H
import Haskoin.Crypto.Keys.Extended
import ZcashHaskell.Types

--  ( AccountId
--  , CoinType(..)
--  , Scope(..)
--  , Seed(..)
--  , ToBytes(..)
--  , TransparentAddress(..)
--  , TransparentReceiver(..)
--  , TransparentSpendingKey(..)
--  , TransparentType(..)
--  , ZcashNet(..)
--  , getTransparentPrefix
--  , getValue
--  )
import ZcashHaskell.Utils (decodeBech32, encodeBech32m)

-- | Required for `TransparentReceiver` encoding and decoding 
sha256 :: BS.ByteString -> BS.ByteString
sha256 bs = BA.convert (hash bs :: Digest SHA256)

-- | Encodes a `TransparentReceiver` into the human-readable format per the Zcash Protocol section 5.6.1.1
encodeTransparentReceiver ::
     ZcashNet -- ^ The network, `MainNet` or `TestNet`
  -> TransparentReceiver -- ^ The address to encode
  -> T.Text
encodeTransparentReceiver zNet t =
  encodeTransparent' (getTransparentPrefix zNet (tr_type t)) $
  toBytes $ tr_bytes t
  where
    encodeTransparent' :: (Word8, Word8) -> BS.ByteString -> T.Text
    encodeTransparent' (a, b) h =
      E.decodeUtf8 $ encodeBase58 bitcoinAlphabet $ digest <> BS.take 4 checksum
      where
        digest = BS.pack [a, b] <> h
        checksum = sha256 $ sha256 digest

-- | Generate an Extended Private Key from a known HDSeed.
genTransparentPrvKey ::
     Seed -- ^ The cryptographic seed of the wallet
  -> CoinType -- ^ The coin type constant to be used
  -> AccountId -- ^ The index of the account to be used
  -> IO TransparentSpendingKey
genTransparentPrvKey hdseed ctype accid = do
  let coin = getValue ctype
  ioCtx <- createContext
  let path = Deriv :| 44 :| coin :| fromIntegral accid :: DerivPath
  let prvKey = makeXPrvKey $ getBytes hdseed
  return $ derivePath ioCtx path prvKey

-- | Generate a transparent receiver
genTransparentReceiver ::
     Int -- ^ The index of the address to be created
  -> Scope -- ^ `External` for wallet addresses or `Internal` for change addresses
  -> XPrvKey -- ^ The transparent private key
  -> IO TransparentReceiver
genTransparentReceiver i scope xprvk = do
  ioCtx <- createContext
  let s =
        case scope of
          External -> 0
          Internal -> 1
  let path = Deriv :/ s :/ fromIntegral i :: DerivPath
  let childPrvKey = derivePath ioCtx path xprvk
  let childPubKey = deriveXPubKey ioCtx childPrvKey
  let x = xPubAddr ioCtx childPubKey
  case x of
    PubKeyAddress k -> return $ TransparentReceiver P2PKH $ fromBinary k
    ScriptAddress j -> return $ TransparentReceiver P2SH $ fromBinary j
    _anyOtherKind -> throwIO $ userError "Unsupported transparent address type"

-- | Generate a transparent receiver
genTransparentSecretKey ::
     Int -- ^ The index of the address to be created
  -> Scope -- ^ `External` for wallet addresses or `Internal` for change addresses
  -> XPrvKey -- ^ The transparent private key
  -> IO TransparentSpendingKey
genTransparentSecretKey i scope xprvk = do
  ioCtx <- createContext
  let s =
        case scope of
          External -> 0
          Internal -> 1
  let path = Deriv :/ s :/ fromIntegral i :: DerivPath
  return $ derivePath ioCtx path xprvk

-- | decode a Transparent Address in HRF and return a TransparentAddress object
decodeTransparentAddress :: BS.ByteString -> Maybe TransparentAddress
decodeTransparentAddress taddress = do
  if BS.length taddress < 34
    then Nothing -- Not a valid transparent address
    else do
      let maybeDecoded = decodeBase58 bitcoinAlphabet taddress
      case maybeDecoded of
        Nothing -> Nothing
        Just decoded -> do
          let digest = BS.take 22 decoded
          let chksum = BS.drop 22 decoded
          let chksumd = BS.take 4 (sha256 $ sha256 digest)
          if chksum /= chksumd
            then Nothing -- Invalid address ( invalid checksum )
              -- build the TransparentAddress Object
            else do
              let addressType = BS.take 2 digest
              let transparentReceiver = BS.drop 2 digest
              let fb = BS.index addressType 0
              let sb = BS.index addressType 1
              case fb of
                28 ->
                  case sb of
                    189 ->
                      Just $
                      TransparentAddress MainNet $
                      TransparentReceiver
                        P2SH
                        (fromRawBytes transparentReceiver)
                    186 ->
                      Just $
                      TransparentAddress TestNet $
                      TransparentReceiver
                        P2SH
                        (fromRawBytes transparentReceiver)
                    184 ->
                      Just $
                      TransparentAddress MainNet $
                      TransparentReceiver
                        P2PKH
                        (fromRawBytes transparentReceiver)
                    _ -> Nothing
                29 ->
                  if sb == 37
                    then Just $
                         TransparentAddress TestNet $
                         TransparentReceiver
                           P2PKH
                           (fromRawBytes transparentReceiver)
                    else Nothing
                _ -> Nothing

-- | Encode an Exchange Addresss into HRF from TransparentReceiver
encodeExchangeAddress :: ZcashNet -> TransparentReceiver -> Maybe T.Text
encodeExchangeAddress net tr = do
  case (tr_type tr) of
    P2PKH -> do
      case net of
        MainNet -> do
          let vhash = encodeBech32m (BC.pack "tex") (toBytes (tr_bytes tr))
          Just vhash
        TestNet -> do
          let vhash = encodeBech32m (BC.pack "textest") (toBytes (tr_bytes tr))
          Just vhash
    _ -> Nothing

-- | Decode an Exchange Address into a ExchangeAddress
decodeExchangeAddress :: T.Text -> Maybe ExchangeAddress
decodeExchangeAddress ex = do
  if (T.length ex) > 1
    then do
      let rawd = decodeBech32 (E.encodeUtf8 ex)
      let tMain = BS.unpack (BC.pack "tex")
      let tTest = BS.unpack (BC.pack "textest")
      let tFail = BS.unpack (BC.pack "fail")
      let hr = BS.unpack (hrp rawd)
      if hr /= tFail
        then do
          let transparentReceiver = bytes rawd
          if hr == tMain
            then Just $
                 ExchangeAddress MainNet $
                 TransparentReceiver P2PKH (fromRawBytes transparentReceiver)
            else do
              if hr == tTest
                then Just $
                     ExchangeAddress TestNet $
                     TransparentReceiver
                       P2PKH
                       (fromRawBytes transparentReceiver)
                else Nothing
        else Nothing
    else Nothing
