-- Copyright 2022-2024 Vergara Technologies LLC
-- 
-- This file is part of Zcash-Haskell.
--
-- |
-- Module      :  ZcashHaskell.Types
-- Copyright   :  2022-2024 Vergara Technologies
-- License     :  MIT
--
-- Maintainer  :  pitmutt@vergara.tech
-- Stability   :  experimental
-- Portability :  unknown
--
-- The types used by the ZcashHaskell library to interact with the Zcash blockchain
--
{-# LANGUAGE DeriveGeneric #-}
{-# LANGUAGE DuplicateRecordFields #-}
{-# LANGUAGE GeneralisedNewtypeDeriving #-}
{-# LANGUAGE StandaloneDeriving #-}
{-# LANGUAGE DeriveAnyClass #-}
{-# LANGUAGE DerivingVia #-}
{-# LANGUAGE UndecidableInstances #-}
{-# LANGUAGE OverloadedStrings #-}

module ZcashHaskell.Types where

import Codec.Borsh
import Crypto.Hash
import Data.Aeson
import qualified Data.ByteArray as BA
import qualified Data.ByteString as BS
import qualified Data.ByteString.Char8 as C
import qualified Data.ByteString.Lazy.UTF8 as US
import qualified Data.ByteString.Short as BS (ShortByteString, toShort)
import Data.HexString
import Data.Int
import Data.Maybe (fromJust, fromMaybe)
import Data.Structured
import qualified Data.Text as T
import qualified Data.Text.Encoding as E
import Data.Word
import qualified GHC.Generics as GHC
import qualified Generics.SOP as SOP
import Haskoin.Address (Address)
import qualified Haskoin.Crypto.Hash as H (Hash256(..))
import Haskoin.Crypto.Keys.Extended (XPrvKey)
import qualified Haskoin.Transaction.Common as H

-- * General
--
-- | A seed for generating private keys
newtype Seed =
  Seed BS.ByteString
  deriving stock (Eq, Prelude.Show, GHC.Generic)
  deriving anyclass (SOP.Generic, SOP.HasDatatypeInfo)
  deriving anyclass (Data.Structured.Show)
  deriving (BorshSize, ToBorsh, FromBorsh) via AsStruct Seed

instance ToBytes Seed where
  getBytes (Seed x) = x

-- | A mnemonic phrase used to derive seeds
newtype Phrase =
  Phrase C.ByteString
  deriving stock (Eq, Prelude.Show, GHC.Generic, Read)
  deriving anyclass (SOP.Generic, SOP.HasDatatypeInfo)
  deriving anyclass (Data.Structured.Show)
  deriving (BorshSize, ToBorsh, FromBorsh) via AsStruct Phrase

instance ToBytes Phrase where
  getBytes (Phrase x) = x

-- | Scope for addresses/receivers
data Scope
  = External -- ^ Addresses used publically to receive payments
  | Internal -- ^ Addresses used internally by wallets for change and shielding
  deriving (Eq, Prelude.Show, Read)

-- | Type to represent data after Bech32 decoding
data RawData = RawData
  { hrp :: !BS.ByteString -- ^ Human-readable part of the Bech32 encoding
  , bytes :: !BS.ByteString -- ^ Decoded bytes
  } deriving stock (Prelude.Show, GHC.Generic)
    deriving anyclass (SOP.Generic, SOP.HasDatatypeInfo)
    deriving anyclass (Data.Structured.Show)
    deriving (BorshSize, ToBorsh, FromBorsh) via AsStruct RawData

-- | Type for the different networks of the Zcash blockchain
data ZcashNet
  = MainNet
  | TestNet
  | RegTestNet
  deriving (Eq, Prelude.Show, Read)

type AccountId = Int

-- | Function to get the Base58 prefix for encoding a 'TransparentReceiver'
getTransparentPrefix :: ZcashNet -> TransparentType -> (Word8, Word8)
getTransparentPrefix n t =
  case t of
    P2SH ->
      case n of
        MainNet -> (0x1c, 0xbd)
        _ -> (0x1c, 0xba)
    P2PKH ->
      case n of
        MainNet -> (0x1c, 0xb8)
        _ -> (0x1d, 0x25)

-- ** Constants
-- | Type for coin types on the different networks
data CoinType
  = MainNetCoin
  | TestNetCoin
  | RegTestNetCoin
  deriving (Eq, Prelude.Show, Ord)

getValue :: CoinType -> Word32
getValue c =
  case c of
    MainNetCoin -> 133
    TestNetCoin -> 1
    RegTestNetCoin -> 1

-- | A Zcash transaction
data Transaction = Transaction
  { tx_id :: !HexString
  , tx_height :: !Int
  , tx_conf :: !Int
  , tx_expiry :: !Int
  , tx_transpBundle :: !(Maybe TransparentBundle)
  , tx_saplingBundle :: !(Maybe SaplingBundle)
  , tx_orchardBundle :: !(Maybe OrchardBundle)
  } deriving (Prelude.Show, Eq, Read)

-- | The transparent portion of a Zcash transaction
data TransparentBundle = TransparentBundle
  { tb_vin :: ![H.TxIn]
  , tb_vout :: ![H.TxOut]
  , tb_coinbase :: !Bool
  } deriving (Eq, Prelude.Show, Read)

-- | Read a raw transparent bundle into the Haskell type
fromRawTBundle :: RawTBundle -> Maybe TransparentBundle
fromRawTBundle rtb =
  if ztb_empty rtb
    then Nothing
    else Just $
         TransparentBundle
           (map fromRawTxIn $ ztb_vin rtb)
           (map fromRawTxOut $ ztb_vout rtb)
           (ztb_coinbase rtb)

fromRawTxIn :: RawTxIn -> H.TxIn
fromRawTxIn t = H.TxIn op (rti_script t) (rti_seq t)
  where
    op =
      if rop_hash (rti_outpoint t) ==
         "\NUL\NUL\NUL\NUL\NUL\NUL\NUL\NUL\NUL\NUL\NUL\NUL\NUL\NUL\NUL\NUL\NUL\NUL\NUL\NUL\NUL\NUL\NUL\NUL\NUL\NUL\NUL\NUL\NUL\NUL\NUL\NUL"
        then H.nullOutPoint
        else H.OutPoint
               ((fromJust .
                 H.hexToTxHash . toText . fromRawBytes . rop_hash . rti_outpoint)
                  t)
               (rop_n $ rti_outpoint t)

fromRawTxOut :: RawTxOut -> H.TxOut
fromRawTxOut t = H.TxOut (rto_amt t) (rto_script t)

-- *** Constants for Sapling Human-readable part
sapExtSpendingKeyHrp = "secret-extended-key-main" :: String

sapExtFullViewingKeyHrp = "zxviews" :: String

sapPaymentAddressHrp = "zs" :: String

sapTestExtSpendingKeyHrp = "secret-extended-key-test" :: String

sapTestExtFullViewingKeyHrp = "zxviewtestsapling" :: String

sapTestPaymentAddressHrp = "ztestsapling" :: String

-- *** Constants for Unified Human-readable part
uniPaymentAddressHrp = "u" :: T.Text

uniFullViewingKeyHrp = "uview" :: T.Text

uniIncomingViewingKeyHrp = "uivk" :: T.Text

uniTestPaymentAddressHrp = "utest" :: T.Text

uniTestFullViewingKeyHrp = "uviewtest" :: T.Text

uniTestIncomingViewingKeyHrp = "uivktest" :: T.Text

-- * RPC
-- | A type to model Zcash RPC calls
data RpcCall = RpcCall
  { jsonrpc :: !T.Text
  , callId :: !T.Text
  , method :: !T.Text
  , parameters :: ![Data.Aeson.Value]
  } deriving stock (Prelude.Show, GHC.Generic)

instance ToJSON RpcCall where
  toJSON (RpcCall j c m p) =
    object ["jsonrpc" .= j, "id" .= c, "method" .= m, "params" .= p]

-- | A type to model the response of the Zcash RPC
data RpcResponse r = MakeRpcResponse
  { err :: !(Maybe RpcError)
  , respId :: !T.Text
  , result :: !(Maybe r)
  } deriving stock (Prelude.Show, GHC.Generic)
    deriving anyclass (ToJSON)

instance (FromJSON r) => FromJSON (RpcResponse r) where
  parseJSON =
    withObject "RpcResponse" $ \obj -> do
      e <- obj .:? "error"
      i <- obj .: "id"
      r <- obj .:? "result"
      pure $ MakeRpcResponse e i r

-- | A type to model the errors from the Zcash RPC
data RpcError = RpcError
  { ecode :: !Double
  , emessage :: !T.Text
  } deriving stock (Prelude.Show, GHC.Generic)
    deriving anyclass (ToJSON)

instance FromJSON RpcError where
  parseJSON =
    withObject "RpcError" $ \obj -> do
      c <- obj .: "code"
      m <- obj .: "message"
      pure $ RpcError c m

-- ** `zcashd`
-- | Type to represent response from the `zcashd` RPC `getblock` method
data BlockResponse = BlockResponse
  { bl_confirmations :: !Integer -- ^ Block confirmations
  , bl_height :: !Integer -- ^ Block height
  , bl_time :: !Integer -- ^ Block time
  , bl_txs :: ![HexString] -- ^ List of transaction IDs in the block
  } deriving (Prelude.Show, Eq)

instance FromJSON BlockResponse where
  parseJSON =
    withObject "BlockResponse" $ \obj -> do
      c <- obj .: "confirmations"
      h <- obj .: "height"
      t <- obj .:? "time"
      txs <- obj .: "tx"
      pure $ BlockResponse c h (fromMaybe 0 t) txs

-- | Type to represent response from the `zcashd` RPC `getrawtransaction`
data RawTxResponse = RawTxResponse
  { rt_id :: !HexString
  , rt_hex :: !HexString
  , rt_shieldedSpends :: ![ShieldedSpend]
  , rt_shieldedOutputs :: ![ShieldedOutput]
  , rt_orchardActions :: ![OrchardAction]
  , rt_blockheight :: !Integer
  , rt_confirmations :: !Integer
  , rt_blocktime :: !Integer
  } deriving (Prelude.Show, Eq, Read)

-- ** `zebrad`
data ZebraTxResponse = ZebraTxResponse
  { ztr_blockheight :: !Int
  , ztr_conf :: !Int
  , ztr_hex :: !HexString
  } deriving (Prelude.Show, Eq, Read)

instance FromJSON ZebraTxResponse where
  parseJSON =
    withObject "ZebraTxResponse" $ \obj -> do
      hex <- obj .: "hex"
      height <- obj .: "height"
      c <- obj .: "confirmations"
      pure $ ZebraTxResponse height c hex

-- | Type to represent a raw deserialized Zebra transaction
data RawZebraTx = RawZebraTx
  { zt_id :: !HexString
  , zt_locktime :: !Word32
  , zt_expiry :: !Word32
  , zt_tBundle :: !RawTBundle
  , zt_sBundle :: !RawSBundle
  , zt_oBundle :: !RawOBundle
  } deriving stock (Eq, Prelude.Show, GHC.Generic)
    deriving anyclass (SOP.Generic, SOP.HasDatatypeInfo)
    deriving anyclass (Data.Structured.Show)
    deriving (BorshSize, ToBorsh, FromBorsh) via AsStruct RawZebraTx

-- | Type for a raw deserialized Zebra transparent bundle
data RawTBundle = RawTBundle
  { ztb_empty :: !Bool
  , ztb_vin :: ![RawTxIn]
  , ztb_vout :: ![RawTxOut]
  , ztb_coinbase :: !Bool
  } deriving stock (Eq, Prelude.Show, GHC.Generic)
    deriving anyclass (SOP.Generic, SOP.HasDatatypeInfo)
    deriving anyclass (Data.Structured.Show)
    deriving (BorshSize, ToBorsh, FromBorsh) via AsStruct RawTBundle

-- | Type for a raw deserialized Zebra Sapling bundle
data RawSBundle = RawSBundle
  { zsb_empty :: !Bool
  , zsb_spends :: ![ShieldedSpend]
  , zsb_outputs :: ![ShieldedOutput]
  , zsb_value :: !Int64
  , zsb_sig :: !HexString
  } deriving stock (Eq, Prelude.Show, GHC.Generic)
    deriving anyclass (SOP.Generic, SOP.HasDatatypeInfo)
    deriving anyclass (Data.Structured.Show)
    deriving (BorshSize, ToBorsh, FromBorsh) via AsStruct RawSBundle

data SaplingBundle = SaplingBundle
  { sbSpends :: ![ShieldedSpend]
  , sbOutputs :: ![ShieldedOutput]
  , sbValue :: !Int64
  , sbSig :: !HexString
  } deriving stock (Eq, Prelude.Show, GHC.Generic, Read)

fromRawSBundle :: RawSBundle -> Maybe SaplingBundle
fromRawSBundle b =
  if zsb_empty b
    then Nothing
    else Just $
         SaplingBundle (zsb_spends b) (zsb_outputs b) (zsb_value b) (zsb_sig b)

-- | Type for a raw deseralized Zebra Orchard bundle
data RawOBundle = RawOBundle
  { zob_empty :: !Bool
  , zob_actions :: ![OrchardAction]
  , zob_flags :: !OrchardFlags
  , zob_value :: !Int64
  , zob_anchor :: !HexString
  , zob_proof :: !HexString
  , zob_sig :: !HexString
  } deriving stock (Eq, Prelude.Show, GHC.Generic)
    deriving anyclass (SOP.Generic, SOP.HasDatatypeInfo)
    deriving anyclass (Data.Structured.Show)
    deriving (BorshSize, ToBorsh, FromBorsh) via AsStruct RawOBundle

-- | Type for an Orchard Bundle
data OrchardBundle = OrchardBundle
  { obActions :: ![OrchardAction]
  , obFlags :: !OrchardFlags
  , obValue :: !Int64
  , obAnchor :: !HexString
  , obProof :: !HexString
  , obSig :: !HexString
  } deriving stock (Eq, Prelude.Show, GHC.Generic, Read)

fromRawOBundle :: RawOBundle -> Maybe OrchardBundle
fromRawOBundle b =
  if zob_empty b
    then Nothing
    else Just $
         OrchardBundle
           (zob_actions b)
           (zob_flags b)
           (zob_value b)
           (zob_anchor b)
           (zob_proof b)
           (zob_sig b)

-- | Type for the Orchard bundle flags
data OrchardFlags = OrchardFlags
  { of_spends :: !Bool
  , of_outputs :: !Bool
  } deriving stock (Eq, Prelude.Show, GHC.Generic, Read)
    deriving anyclass (SOP.Generic, SOP.HasDatatypeInfo)
    deriving anyclass (Data.Structured.Show)
    deriving (BorshSize, ToBorsh, FromBorsh) via AsStruct OrchardFlags

-- | Type for the response from the `zebrad` RPC method `getinfo`
data ZebraGetInfo = ZebraGetInfo
  { zgi_build :: !T.Text
  , zgi_subversion :: !T.Text
  } deriving (Prelude.Show, Eq)

instance FromJSON ZebraGetInfo where
  parseJSON =
    withObject "ZebraGetInfo" $ \obj -> do
      b <- obj .: "build"
      s <- obj .: "subversion"
      pure $ ZebraGetInfo b s

-- | Type for the response from the `zebrad` RPC method `getblockchaininfo`
data ZebraGetBlockChainInfo = ZebraGetBlockChainInfo
  { zgb_best :: !HexString
  , zgb_blocks :: !Int
  , zgb_net :: !ZcashNet
  } deriving (Prelude.Show, Eq)

instance FromJSON ZebraGetBlockChainInfo where
  parseJSON =
    withObject "ZebraGetBlockChainInfo" $ \obj -> do
      be <- obj .: "bestblockhash"
      b <- obj .: "blocks"
      c <- obj .: "chain"
      pure $
        ZebraGetBlockChainInfo
          be
          b
          (case (c :: String) of
             "main" -> MainNet
             "test" -> TestNet
             _ -> RegTestNet)

-- * Transparent
-- | Type to represent the two kinds of transparent addresses
data TransparentType
  = P2SH
  | P2PKH
  deriving (Eq, Prelude.Show, Read)

-- | Type for transparent spending key
type TransparentSpendingKey = XPrvKey

-- | Type to represent a transparent Zcash addresses
data TransparentReceiver = TransparentReceiver
  { tr_type :: !TransparentType
  , tr_bytes :: !HexString
  } deriving (Eq, Prelude.Show, Read)

-- | Type to represent a transparent Zcash addresses
data TransparentAddress = TransparentAddress
  { ta_network :: !ZcashNet
  , ta_receiver :: !TransparentReceiver
  } deriving (Eq, Prelude.Show, Read)

-- | Type to represent a TEX Zcash addresses
data ExchangeAddress = ExchangeAddress
  { ex_network :: !ZcashNet
  , ex_address :: !TransparentReceiver
  } deriving (Eq, Prelude.Show, Read)

-- | Wrapper types for transparent elements
data RawTxIn = RawTxIn
  { rti_outpoint :: !RawOutPoint
  , rti_script :: !BS.ByteString
  , rti_seq :: !Word32
  } deriving stock (Eq, Prelude.Show, GHC.Generic)
    deriving anyclass (SOP.Generic, SOP.HasDatatypeInfo)
    deriving anyclass (Data.Structured.Show)
    deriving (BorshSize, ToBorsh, FromBorsh) via AsStruct RawTxIn

data RawTxOut = RawTxOut
  { rto_amt :: !Word64
  , rto_script :: !BS.ByteString
  } deriving stock (Eq, Prelude.Show, GHC.Generic)
    deriving anyclass (SOP.Generic, SOP.HasDatatypeInfo)
    deriving anyclass (Data.Structured.Show)
    deriving (BorshSize, ToBorsh, FromBorsh) via AsStruct RawTxOut

data RawOutPoint = RawOutPoint
  { rop_hash :: !BS.ByteString
  , rop_n :: !Word32
  } deriving stock (Eq, Prelude.Show, GHC.Generic)
    deriving anyclass (SOP.Generic, SOP.HasDatatypeInfo)
    deriving anyclass (Data.Structured.Show)
    deriving (BorshSize, ToBorsh, FromBorsh) via AsStruct RawOutPoint

-- * Sapling
-- | A spending key for Sapling
newtype SaplingSpendingKey =
  SaplingSpendingKey BS.ByteString
  deriving stock (Eq, Prelude.Show, Read)

instance ToBytes SaplingSpendingKey where
  getBytes (SaplingSpendingKey s) = s

-- | A Sapling receiver
newtype SaplingReceiver =
  SaplingReceiver BS.ByteString
  deriving stock (Eq, Prelude.Show, Read)

instance ToBytes SaplingReceiver where
  getBytes (SaplingReceiver s) = s

data SaplingAddress = SaplingAddress
  { net_type :: !ZcashNet
  , sa_receiver :: !SaplingReceiver
  } deriving (Eq, Prelude.Show, Read)

-- | Type to represent a Sapling Shielded Spend as provided by the @getrawtransaction@ RPC method
data ShieldedSpend = ShieldedSpend
  { sp_cv :: !HexString
  , sp_anchor :: !HexString
  , sp_nullifier :: !HexString
  , sp_rk :: !HexString
  , sp_proof :: !HexString
  , sp_auth :: !HexString
  } deriving stock (Eq, Prelude.Show, GHC.Generic, Read)
    deriving anyclass (SOP.Generic, SOP.HasDatatypeInfo)
    deriving anyclass (Data.Structured.Show)
    deriving (BorshSize, ToBorsh, FromBorsh) via AsStruct ShieldedSpend

instance FromJSON ShieldedSpend where
  parseJSON =
    withObject "ShieldedSpend" $ \obj -> do
      cv <- obj .: "cv"
      anchor <- obj .: "anchor"
      nullifier <- obj .: "nullifier"
      rk <- obj .: "rk"
      p <- obj .: "proof"
      sig <- obj .: "spendAuthSig"
      pure $ ShieldedSpend cv anchor nullifier rk p sig

-- | Type to represent a Sapling Shielded Output as provided by the @getrawtransaction@ RPC method of @zcashd@.
data ShieldedOutput = ShieldedOutput
  { s_cv :: !HexString -- ^ Value commitment to the input note
  , s_cmu :: !HexString -- ^ The u-coordinate of the note commitment for the output note
  , s_ephKey :: !HexString -- ^ Ephemeral Jubjub public key
  , s_encCipherText :: !HexString -- ^ The output note encrypted to the recipient
  , s_outCipherText :: !HexString -- ^ A ciphertext enabling the sender to recover the output note
  , s_proof :: !HexString -- ^ Zero-knowledge proof using the Sapling Output circuit
  } deriving stock (Eq, Prelude.Show, GHC.Generic, Read)
    deriving anyclass (SOP.Generic, SOP.HasDatatypeInfo)
    deriving anyclass (Data.Structured.Show)
    deriving (BorshSize, ToBorsh, FromBorsh) via AsStruct ShieldedOutput

instance FromJSON ShieldedOutput where
  parseJSON =
    withObject "ShieldedOutput" $ \obj -> do
      cv <- obj .: "cv"
      cmu <- obj .: "cmu"
      ephKey <- obj .: "ephemeralKey"
      encText <- obj .: "encCiphertext"
      outText <- obj .: "outCiphertext"
      p <- obj .: "proof"
      pure $ ShieldedOutput cv cmu ephKey encText outText p

-- | Type for a Sapling note commitment tree
newtype SaplingCommitmentTree = SaplingCommitmentTree
  { sapTree :: HexString
  } deriving (Eq, Prelude.Show, Read)

-- | Type for a Sapling incremental witness
newtype SaplingWitness = SaplingWitness
  { sapWit :: HexString
  } deriving (Eq, Prelude.Show, Read)

-- * Orchard
-- | A spending key for Orchard
newtype OrchardSpendingKey =
  OrchardSpendingKey BS.ByteString
  deriving stock (Eq, Prelude.Show, Read)

instance ToBytes OrchardSpendingKey where
  getBytes (OrchardSpendingKey o) = o

-- | An Orchard receiver
newtype OrchardReceiver =
  OrchardReceiver BS.ByteString
  deriving stock (Eq, Prelude.Show, Read)

instance ToBytes OrchardReceiver where
  getBytes (OrchardReceiver o) = o

-- | Type to represent a Unified Address
data UnifiedAddress = UnifiedAddress
  { ua_net :: !ZcashNet
  , o_rec :: !(Maybe OrchardReceiver)
  , s_rec :: !(Maybe SaplingReceiver)
  , t_rec :: !(Maybe TransparentReceiver)
  } deriving (Prelude.Show, Eq, Read)

-- | Helper type for marshalling UAs 
data RawUA = RawUA
  { raw_net :: !Word8
  , raw_o :: !BS.ByteString
  , raw_s :: !BS.ByteString
  , raw_t :: !BS.ByteString
  , raw_to :: !BS.ByteString
  } deriving stock (Eq, Prelude.Show, GHC.Generic)
    deriving anyclass (SOP.Generic, SOP.HasDatatypeInfo)
    deriving anyclass (Data.Structured.Show)
    deriving (BorshSize, ToBorsh, FromBorsh) via AsStruct RawUA

-- | Type to represent a Unified Full Viewing Key
data UnifiedFullViewingKey = UnifiedFullViewingKey
  { net :: !Word8 -- ^ Number representing the network the key belongs to. @1@ for @mainnet@, @2@ for @testnet@ and @3@ for @regtestnet@.
  , o_key :: !BS.ByteString -- ^ Raw bytes of the Orchard Full Viewing Key as specified in [ZIP-316](https://zips.z.cash/zip-0316)
  , s_key :: !BS.ByteString -- ^ Raw bytes of the Sapling Full Viewing Key as specified in [ZIP-316](https://zips.z.cash/zip-0316)
  , t_key :: !BS.ByteString -- ^ Raw bytes of the P2PKH chain code and public key as specified in [ZIP-316](https://zips.z.cash/zip-0316)
  } deriving stock (Eq, Prelude.Show, GHC.Generic)
    deriving anyclass (SOP.Generic, SOP.HasDatatypeInfo)
    deriving anyclass (Data.Structured.Show)
    deriving (BorshSize, ToBorsh, FromBorsh) via AsStruct UnifiedFullViewingKey

-- | Type to represent an Orchard Action as provided by the @getrawtransaction@ RPC method of @zcashd@, and defined in the [Zcash Protocol](https://zips.z.cash/protocol/protocol.pdf)
data OrchardAction = OrchardAction
  { nf :: !HexString -- ^ The nullifier of the input note
  , rk :: !HexString -- ^ The randomized validating key for @auth@
  , cmx :: !HexString -- ^ The x-coordinate of the note commitment for the output note
  , eph_key :: !HexString -- ^ An encoding of an ephemeral Pallas public key
  , enc_ciphertext :: !HexString -- ^ The output note encrypted to the recipient
  , out_ciphertext :: !HexString -- ^ A ciphertext enabling the sender to recover the output note
  , cv :: !HexString -- ^ A value commitment to the net value of the input note minus the output note
  , auth :: !HexString -- ^ A signature authorizing the spend in this Action
  } deriving stock (Eq, Prelude.Show, GHC.Generic, Read)
    deriving anyclass (SOP.Generic, SOP.HasDatatypeInfo)
    deriving anyclass (Data.Structured.Show)
    deriving (BorshSize, ToBorsh, FromBorsh) via AsStruct OrchardAction

instance FromJSON OrchardAction where
  parseJSON =
    withObject "OrchardAction" $ \obj -> do
      n <- obj .: "nullifier"
      r <- obj .: "rk"
      c <- obj .: "cmx"
      ephKey <- obj .: "ephemeralKey"
      encText <- obj .: "encCiphertext"
      outText <- obj .: "outCiphertext"
      cval <- obj .: "cv"
      a <- obj .: "spendAuthSig"
      pure $ OrchardAction n r c ephKey encText outText cval a

-- | Type for a Orchard note commitment tree
newtype OrchardCommitmentTree = OrchardCommitmentTree
  { orchTree :: HexString
  } deriving (Eq, Prelude.Show, Read)

-- | Type for a Sapling incremental witness
newtype OrchardWitness = OrchardWitness
  { orchWit :: HexString
  } deriving (Eq, Prelude.Show, Read)

-- | Type to represent a decoded note
data DecodedNote = DecodedNote
  { a_value :: !Int64 -- ^ The amount of the transaction in _zatoshis_.
  , a_recipient :: !BS.ByteString -- ^ The recipient Orchard receiver.
  , a_memo :: !BS.ByteString -- ^ The decoded shielded memo field.
  , a_nullifier :: !HexString -- ^ The calculated nullifier
  , a_rho :: !BS.ByteString
  , a_rseed :: !Rseed
  } deriving stock (Eq, Prelude.Show, GHC.Generic)
    deriving anyclass (SOP.Generic, SOP.HasDatatypeInfo)
    deriving anyclass (Data.Structured.Show)
    deriving (BorshSize, ToBorsh, FromBorsh) via AsStruct DecodedNote

data Rseed = Rseed
  { rs_kind :: !Word8
  , rs_bytes :: !BS.ByteString
  } deriving stock (Eq, Prelude.Show, GHC.Generic, Read)
    deriving anyclass (SOP.Generic, SOP.HasDatatypeInfo)
    deriving anyclass (Data.Structured.Show)
    deriving (BorshSize, ToBorsh, FromBorsh) via AsStruct Rseed

data TransparentTxSpend = TransparentTxSpend
  { ts_sk :: !BS.ByteString
  , ts_utxo :: !RawOutPoint
  , ts_coin :: !RawTxOut
  } deriving stock (Eq, Prelude.Show, GHC.Generic)
    deriving anyclass (SOP.Generic, SOP.HasDatatypeInfo)
    deriving anyclass (Data.Structured.Show)
    deriving (BorshSize, ToBorsh, FromBorsh) via AsStruct TransparentTxSpend

data SaplingTxSpend = SaplingTxSpend
  { ss_sk :: !BS.ByteString
  , ss_note :: !DecodedNote
  , ss_iw :: !BS.ByteString
  } deriving stock (Eq, Prelude.Show, GHC.Generic)
    deriving anyclass (SOP.Generic, SOP.HasDatatypeInfo)
    deriving anyclass (Data.Structured.Show)
    deriving (BorshSize, ToBorsh, FromBorsh) via AsStruct SaplingTxSpend

data OrchardTxSpend = OrchardTxSpend
  { ss_sk :: !BS.ByteString
  , ss_note :: !DecodedNote
  , ss_iw :: !BS.ByteString
  } deriving stock (Eq, Prelude.Show, GHC.Generic)
    deriving anyclass (SOP.Generic, SOP.HasDatatypeInfo)
    deriving anyclass (Data.Structured.Show)
    deriving (BorshSize, ToBorsh, FromBorsh) via AsStruct OrchardTxSpend

data OutgoingNote = OutgoingNote
  { on_kind :: !Word8
  , on_key :: !BS.ByteString
  , on_recipient :: !BS.ByteString
  , on_amt :: !Word64
  , on_memo :: !BS.ByteString
  , on_chg :: !Bool
  } deriving stock (Eq, Prelude.Show, GHC.Generic)
    deriving anyclass (SOP.Generic, SOP.HasDatatypeInfo)
    deriving anyclass (Data.Structured.Show)
    deriving (BorshSize, ToBorsh, FromBorsh) via AsStruct OutgoingNote

newtype SaplingSpendParams = SaplingSpendParams
  { sapSParams :: BS.ByteString
  } deriving newtype (Eq, Prelude.Show, Read)

newtype SaplingOutputParams = SaplingOutputParams
  { sapOParams :: BS.ByteString
  } deriving newtype (Eq, Prelude.Show, Read)

data TxError
  = InsufficientFunds
  | ChangeRequired
  | Fee
  | Balance
  | TransparentBuild
  | SaplingBuild
  | OrchardBuild
  | OrchardSpend
  | OrchardRecipient
  | SaplingBuilderNotAvailable
  | OrchardBuilderNotAvailable
  | ZHError
  deriving (Eq, Prelude.Show, Read)

-- * Classes
-- | Class to represent types with a bytestring representation
class ToBytes a where
  getBytes :: a -> BS.ByteString

-- * Helpers
-- | Helper function to turn a hex-encoded string to bytestring
decodeHexText :: String -> BS.ByteString
decodeHexText h = BS.pack $ hexRead h
  where
    hexRead hexText
      | null chunk = []
      | otherwise =
        fromIntegral (read ("0x" <> chunk)) : hexRead (drop 2 hexText)
      where
        chunk = take 2 hexText
