-- Copyright 2022-2024 Vergara Technologies LLC
-- 
-- This file is part of Zcash-Haskell.
--
-- |
-- Module      :  ZcashHaskell.Utils
-- Copyright   :  2022-2024 Vergara Technologies LLC
-- License     :  MIT
--
-- Maintainer  :  pitmutt@vergara.tech
-- Stability   :  experimental
-- Portability :  unknown
--
-- A set of functions to assist in the handling of elements of the Zcash protocol, allowing for decoding of memos, addresses and viewing keys.
--
{-# LANGUAGE OverloadedStrings #-}

module ZcashHaskell.Utils where

import C.Zcash
  ( rustWrapperBech32Decode
  , rustWrapperBech32Encode
  , rustWrapperBech32mEncode
  , rustWrapperCreateTx
  , rustWrapperF4Jumble
  , rustWrapperF4UnJumble
  , rustWrapperTxRead
  )
import Control.Exception (SomeException(..), try)
import Control.Monad.IO.Class
import Data.Aeson
import Data.Binary.Get
import qualified Data.ByteString as BS
import qualified Data.ByteString.Lazy as LBS
import Data.HexString (HexString(..), toBytes)
import qualified Data.Text as T
import qualified Data.Text.Encoding as E
import Foreign.Rust.Marshall.Variable
import Network.HTTP.Client (HttpException(..))
import Network.HTTP.Simple
import ZcashHaskell.Types

-- * Utility functions
-- | Decode the given bytestring using Bech32
decodeBech32 :: BS.ByteString -> RawData
decodeBech32 = withPureBorshVarBuffer . rustWrapperBech32Decode

-- | Encode the given Human Readable Part and bytestring as a Bech32m string
encodeBech32m :: BS.ByteString -> BS.ByteString -> T.Text
encodeBech32m h d = withPureBorshVarBuffer $ rustWrapperBech32mEncode h d

-- | Encode the given Human Readable Part and bytestring as a Bech32 string
encodeBech32 :: BS.ByteString -> BS.ByteString -> T.Text
encodeBech32 h d = withPureBorshVarBuffer $ rustWrapperBech32Encode h d

-- | Apply the F4Jumble transformation to the given bytestring
f4Jumble :: BS.ByteString -> BS.ByteString
f4Jumble = withPureBorshVarBuffer . rustWrapperF4Jumble

-- | Apply the inverse F4Jumble transformation to the given bytestring
f4UnJumble :: BS.ByteString -> BS.ByteString
f4UnJumble = withPureBorshVarBuffer . rustWrapperF4UnJumble

-- * Node interaction
-- | Make a Zcash RPC call
makeZcashCall ::
     (MonadIO m, FromJSON a)
  => BS.ByteString
  -> BS.ByteString
  -> T.Text
  -> [Data.Aeson.Value]
  -> m (Response a)
makeZcashCall username password m p = do
  let payload = RpcCall "1.0" "test" m p
  let myRequest =
        setRequestBodyJSON payload $
        setRequestPort 8232 $
        setRequestBasicAuth username password $
        setRequestMethod "POST" defaultRequest
  httpJSON myRequest

-- | Make a Zebra RPC call
makeZebraCall ::
     FromJSON a
  => T.Text -- ^ Hostname for `zebrad`
  -> Int -- ^ Port for `zebrad`
  -> T.Text -- ^ RPC method to call
  -> [Data.Aeson.Value] -- ^ List of parameters
  -> IO (Either String a)
makeZebraCall host port m params = do
  let payload = RpcCall "2.0" "zh" m params
  let myRequest =
        setRequestBodyJSON payload $
        setRequestPort port $
        setRequestHost (E.encodeUtf8 host) $
        setRequestMethod "POST" defaultRequest
  r <-
    try $ httpJSON myRequest :: FromJSON a1 =>
      IO (Either SomeException (Response (RpcResponse a1)))
  case r of
    Left ex -> return $ Left $ show ex
    Right res -> do
      let zebraResp = getResponseBody res
      case err zebraResp of
        Just zErr -> return $ Left $ T.unpack $ emessage zErr
        Nothing ->
          case result zebraResp of
            Nothing -> return $ Left "Empty response from Zebra"
            Just zR -> return $ Right zR

getBlockTime :: HexString -> Int
getBlockTime hex_block =
  fromIntegral $
  runGet getInt32le $
  LBS.fromStrict $ BS.take 4 (BS.drop 100 $ hexBytes hex_block)

readZebraTransaction :: HexString -> Maybe RawZebraTx
readZebraTransaction hex =
  if BS.length (hexBytes $ zt_id rawTx) < 1
    then Nothing
    else Just rawTx
  where
    rawTx = (withPureBorshVarBuffer . rustWrapperTxRead) $ hexBytes hex

createTransaction ::
     Maybe SaplingCommitmentTree -- ^ to obtain the Sapling anchor
  -> Maybe OrchardCommitmentTree -- ^ to obtain the Orchard anchor
  -> [TransparentTxSpend] -- ^ the list of transparent notes to spend
  -> [SaplingTxSpend] -- ^ the list of Sapling notes to spend
  -> [OrchardTxSpend] -- ^ the list of Orchard notes to spend
  -> [OutgoingNote] -- ^ the list of outgoing notes, including change notes
  -> SaplingSpendParams -- ^ the Sapling circuit spending parameters
  -> SaplingOutputParams -- ^ the Sapling circuit output parameters
  -> ZcashNet -- ^ the network to be used
  -> Int -- ^ target block height
  -> Bool -- ^ True to build, False to estimate fee
  -> Either TxError HexString
createTransaction sapAnchor orchAnchor tSpend sSpend oSpend outgoing sParams oParams znet bh build =
  if BS.length (hexBytes txResult) > 1
    then Right txResult
    else case head (BS.unpack $ hexBytes txResult) of
           0 -> Left InsufficientFunds
           1 -> Left ChangeRequired
           2 -> Left Fee
           3 -> Left Balance
           4 -> Left TransparentBuild
           5 -> Left SaplingBuild
           6 -> Left OrchardBuild
           7 -> Left OrchardSpend
           8 -> Left OrchardRecipient
           9 -> Left SaplingBuilderNotAvailable
           10 -> Left OrchardBuilderNotAvailable
           _ -> Left ZHError
  where
    txResult =
      withPureBorshVarBuffer $
      rustWrapperCreateTx
        (case sapAnchor of
           Nothing -> "0"
           Just sA -> toBytes $ sapTree sA)
        (case orchAnchor of
           Nothing -> "0"
           Just oA -> toBytes $ orchTree oA)
        tSpend
        sSpend
        oSpend
        outgoing
        (sapSParams sParams)
        (sapOParams oParams)
        (znet == MainNet)
        (fromIntegral bh)
        build
